"""Index of XLSForm file paths."""

import os


xlsforms_path = os.path.dirname(os.path.abspath(__file__))

buildings = f"{xlsforms_path}/buildings.xls"
cemeteries = f"{xlsforms_path}/cemeteries.xls"
education = f"{xlsforms_path}/education.xls"
healthcare = f"{xlsforms_path}/healthcare.xls"
landuse = f"{xlsforms_path}/landuse.xls"
natural = f"{xlsforms_path}/natural.xls"
place = f"{xlsforms_path}/place.xls"
religious = f"{xlsforms_path}/religious.xls"
school = f"{xlsforms_path}/school.xls"
solidwaste = f"{xlsforms_path}/solidwaste.xlsx"
toilets = f"{xlsforms_path}/toilets.xls"
transportation = f"{xlsforms_path}/transportation.xls"
waste_collection = f"{xlsforms_path}/waste_collection.xlsx"
waste = f"{xlsforms_path}/waste.xls"
waterpoints = f"{xlsforms_path}/waterpoints.xls"
waterways = f"{xlsforms_path}/waterways.xls"
